/*
 * serve-index
 * Copyright(c) 2011 Sencha Inc.
 * Copyright(c) 2011 TJ Holowaychuk
 * Copyright(c) 2014-2015 Douglas Christopher Wilson
 * https://github.com/expressjs/serve-index
 *
 * serve-directory
 * Copyright(c) 2017- fisker Cheung
 * MIT Licensed
 */

'use strict'

var path = require('path')
var fs = require('fs')
var httpError = require('http-errors')
var accepts = require('accepts')
var parseurl = require('parseurl')
var debug$1 = require('debug')
var url = require('url')
var module$1 = require('module')
var mimeTypes = require('mime-types')
var _template = require('lodash.template')

function _interopDefaultLegacy(e) {
  return e && typeof e === 'object' && 'default' in e ? e : {default: e}
}

var path__default = /*#__PURE__*/ _interopDefaultLegacy(path)
var fs__default = /*#__PURE__*/ _interopDefaultLegacy(fs)
var httpError__default = /*#__PURE__*/ _interopDefaultLegacy(httpError)
var accepts__default = /*#__PURE__*/ _interopDefaultLegacy(accepts)
var debug__default = /*#__PURE__*/ _interopDefaultLegacy(debug$1)
var _template__default = /*#__PURE__*/ _interopDefaultLegacy(_template)

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object)

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object)

    if (enumerableOnly) {
      symbols = symbols.filter(function (sym) {
        return Object.getOwnPropertyDescriptor(object, sym).enumerable
      })
    }

    keys.push.apply(keys, symbols)
  }

  return keys
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {}

    if (i % 2) {
      ownKeys(Object(source), true).forEach(function (key) {
        _defineProperty(target, key, source[key])
      })
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source))
    } else {
      ownKeys(Object(source)).forEach(function (key) {
        Object.defineProperty(
          target,
          key,
          Object.getOwnPropertyDescriptor(source, key)
        )
      })
    }
  }

  return target
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true,
    })
  } else {
    obj[key] = value
  }

  return obj
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return
  if (typeof o === 'string') return _arrayLikeToArray(o, minLen)
  var n = Object.prototype.toString.call(o).slice(8, -1)
  if (n === 'Object' && o.constructor) n = o.constructor.name
  if (n === 'Map' || n === 'Set') return Array.from(o)
  if (n === 'Arguments' || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))
    return _arrayLikeToArray(o, minLen)
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i]

  return arr2
}

function _createForOfIteratorHelper(o, allowArrayLike) {
  var it =
    (typeof Symbol !== 'undefined' && o[Symbol.iterator]) || o['@@iterator']

  if (!it) {
    if (
      Array.isArray(o) ||
      (it = _unsupportedIterableToArray(o)) ||
      (allowArrayLike && o && typeof o.length === 'number')
    ) {
      if (it) o = it
      var i = 0

      var F = function () {}

      return {
        s: F,
        n: function () {
          if (i >= o.length)
            return {
              done: true,
            }
          return {
            done: false,
            value: o[i++],
          }
        },
        e: function (e) {
          throw e
        },
        f: F,
      }
    }

    throw new TypeError(
      'Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.'
    )
  }

  var normalCompletion = true,
    didErr = false,
    err
  return {
    s: function () {
      it = it.call(o)
    },
    n: function () {
      var step = it.next()
      normalCompletion = step.done
      return step
    },
    e: function (e) {
      didErr = true
      err = e
    },
    f: function () {
      try {
        if (!normalCompletion && it.return != null) it.return()
      } finally {
        if (didErr) throw err
      }
    },
  }
}

function create$1(base) {
  return {
    async load(file) {
      const url = new URL(file, base)
      const buffer = await fs.promises.readFile(url)
      return JSON.parse(buffer)
    },
    loadSync(file) {
      const url = new URL(file, base)
      const buffer = fs__default['default'].readFileSync(url)
      return JSON.parse(buffer)
    },
  }
}

function create(importMeta) {
  return {
    get filename() {
      return url.fileURLToPath(importMeta.url)
    },
    get dirname() {
      return path__default['default'].dirname(this.filename)
    },
    get require() {
      return module$1.createRequire(importMeta.url)
    },
    get json() {
      return create$1(importMeta.url)
    },
    // Aliases
    get __filename() {
      return this.filename
    },
    get __dirname() {
      return this.dirname
    },
  }
}

const _createEsmUtils$2 = create({
    url:
      typeof document === 'undefined'
        ? new (require('u' + 'rl').URL)('file:' + __filename).href
        : (document.currentScript && document.currentScript.src) ||
          new URL('index.cjs', document.baseURI).href,
  }),
  json$1 = _createEsmUtils$2.json

const _json$loadSync$1 = json$1.loadSync('../package.json'),
  name$1 = _json$loadSync$1.name

var debug = debug__default['default'](name$1)

function mime(extension) {
  return mimeTypes.lookup(extension) || ''
}

function directoryFirst(a, b) {
  return b.isDirectory() - a.isDirectory()
}

function sortBy(key) {
  return function (a, b) {
    if (b[key] === a[key]) {
      return 0
    }

    return b[key] < a[key] ? 1 : -1
  }
}

function sortFile(a, b) {
  return directoryFirst(a, b) || sortBy('name')(a, b)
}

function startsWithDot(file) {
  return file[0] === '.'
}

function isHiddenFile(file) {
  return startsWithDot(file.name)
}

function notHiddenFile(file) {
  return !isHiddenFile(file)
}

function isHiddenPath(filePath) {
  return path__default['default']
    .normalize(filePath)
    .split(path__default['default'].sep)
    .some((segment) => startsWithDot(segment))
}

class Connection {
  constructor(sd, request, response, next) {
    this.sd = sd
    this.request = request
    this.response = response
    this.next = next
    this.url = parseurl.original(this.request)
  }

  getMethod() {
    const method = this.request.method

    if (method !== 'GET' && method !== 'HEAD') {
      this.response.statusCode = method === 'OPTIONS' ? 200 : 405
      this.response.setHeader('Allow', 'GET, HEAD, OPTIONS')
      this.response.setHeader('Content-Length', '0')
      this.response.end()
      return
    }

    this.method = method
    return method
  }

  getPathname() {
    let pathname = this.url.pathname
    const hidden = this.sd.options.hidden

    try {
      pathname = decodeURIComponent(pathname)
    } catch (_unused) {
      this.next(httpError__default['default'](400))
      return
    }

    if (!hidden && isHiddenPath(pathname)) {
      debug('hidden folder "%s" deny.', pathname)
      this.next(httpError__default['default'](403))
      return
    } // null byte(s), bad request

    if (pathname.includes('\0')) {
      debug('null byte(s) in "%s", bad request.', pathname)
      this.next(httpError__default['default'](400))
      return
    }

    pathname = pathname.replace(/\/+/g, '/')
    this.pathname = pathname
    return pathname
  }

  getResponseType() {
    const acceptMediaTypes = Object.keys(this.sd.responser)
    const responseType = accepts__default['default'](this.request).type(
      acceptMediaTypes
    )

    if (!responseType) {
      debug('mime not acceptable "%s".', responseType)
      this.next(httpError__default['default'](406))
      return
    }

    this.responseType = responseType
    return responseType
  }

  getResponser() {
    const responser = this.sd.responser[this.responseType]
    this.responser = responser
    return responser
  }

  getPath() {
    // join / normalize from root dir
    const filePath = path__default['default'].join(this.sd.root, this.pathname) // malicious path

    if (!filePath.startsWith(this.sd.root + path__default['default'].sep)) {
      debug('malicious path "%s".', this.pathname)
      this.next(httpError__default['default'](403))
      return
    }

    this.path = filePath
    return filePath
  }

  getDirectory() {
    debug('get directory "%s".', this.path)
    let stats

    try {
      stats = fs__default['default'].statSync(this.path)
    } catch (error) {
      if (error.code === 'ENOENT' || error.code === 'ENOTDIR') {
        this.next()
        return
      }

      error.status = error.code === 'ENAMETOOLONG' ? 414 : 500
      this.next(error)
      return
    }

    if (!stats.isDirectory()) {
      this.next()
      return
    }

    if (this.pathname.slice(-1) !== '/') {
      debug('add "/" to "%s".', this.pathname)
      this.response.writeHead(301, {
        Location: `${this.url.pathname}/`,
      })
      this.response.end()
      return
    }

    stats.path = this.path
    stats.pathname = this.pathname
    stats.url = this.sd.options.relative ? '.' : this.url.pathname
    this.directory = stats
    return stats
  }

  getFiles() {
    debug('get files "%s"', this.path)
    const directoryPath = this.path
    const urlPrefix = this.sd.options.relative ? '' : this.url.pathname
    let files

    try {
      files = fs__default['default']
        .readdirSync(directoryPath)
        .map((file) => {
          const stats = fs__default['default'].statSync(
            path__default['default'].join(directoryPath, file)
          )
          stats.name = file
          stats.ext = path__default['default'].extname(file)
          stats.type = mime(stats.ext)
          stats.url =
            urlPrefix +
            encodeURIComponent(file) +
            (stats.isDirectory() ? '/' : '')
          return stats
        })
        .sort(sortFile)
    } catch (error) {
      this.next(error)
      return
    }

    if (!this.sd.options.hidden) {
      files = files.filter((file) => notHiddenFile(file))
    }

    this.files = files
    return files
  }

  process() {
    if (
      !this.getMethod() ||
      !this.getPathname() ||
      !this.getResponseType() ||
      !this.getResponser() ||
      !this.getPath() ||
      !this.getDirectory() ||
      !this.getFiles()
    ) {
      return
    }

    try {
      this.responser(this.request, this.response, {
        path: this.path,
        pathname: this.pathname,
        url: this.url,
        method: this.method,
        responseType: this.responseType,
        directory: this.directory,
        files: this.files,
        fileNames: this.files.map(({name}) => name),
      })
    } catch (error) {
      error.status = 500
      this.next(error)
    }
  }
}

var Connection$1 = Connection

const _createEsmUtils$1 = create({
    url:
      typeof document === 'undefined'
        ? new (require('u' + 'rl').URL)('file:' + __filename).href
        : (document.currentScript && document.currentScript.src) ||
          new URL('index.cjs', document.baseURI).href,
  }),
  __dirname$1 = _createEsmUtils$1.__dirname

const CHARSET = 'utf-8'
const EOL = '\n'
const DEFAULT_HTML_TEMPLATE_FILE = path__default['default'].join(
  __dirname$1,
  'directory.jst'
)

var defaults = {
  hidden: false,
  relative: true,
  process: [
    {
      accept: 'text/html',
      render: DEFAULT_HTML_TEMPLATE_FILE,
    },
    {
      accept: 'text/plain',
      render: ({fileNames}) => fileNames.join(EOL) + EOL,
    },
    {
      accept: 'application/json',
      render: ({fileNames}) => JSON.stringify(fileNames),
    },
  ],
}

function identity(x) {
  return x
}

function isBoolean(x) {
  return x === true || x === false
}

function isFunction(x) {
  return typeof x === 'function'
}

function isString(x) {
  return typeof x === 'string'
}

function noop() {}

function readFile(file) {
  try {
    return fs.readFileSync(file, CHARSET)
  } catch (_unused) {}
}

const toString = Object.prototype.toString

function type(x) {
  return toString.call(x).slice(8, -1)
}

var utils = /*#__PURE__*/ Object.freeze({
  __proto__: null,
  directoryFirst: directoryFirst,
  identity: identity,
  isHiddenFile: isHiddenFile,
  isHiddenPath: isHiddenPath,
  isBoolean: isBoolean,
  isFunction: isFunction,
  isString: isString,
  mime: mime,
  noop: noop,
  notHiddenFile: notHiddenFile,
  readFile: readFile,
  sortBy: sortBy,
  sortFile: sortFile,
  type: type,
})

const _createEsmUtils = create({
    url:
      typeof document === 'undefined'
        ? new (require('u' + 'rl').URL)('file:' + __filename).href
        : (document.currentScript && document.currentScript.src) ||
          new URL('index.cjs', document.baseURI).href,
  }),
  require$1 = _createEsmUtils.require,
  json = _createEsmUtils.json

const _json$loadSync = json.loadSync('../package.json'),
  name = _json$loadSync.name,
  version = _json$loadSync.version

const packageJson = {
  name,
  version,
}
var imports = Object.freeze(
  _objectSpread2(
    {
      CHARSET,
      pkg: packageJson,
      require: require$1,
    },
    utils
  )
)

function template(template, options) {
  if (isFunction(template)) {
    return template
  }

  if (isString(template)) {
    if (fs.existsSync(template)) {
      template = readFile(template)
    }

    return _template__default['default'](
      template,
      options || {
        imports,
      }
    )
  }

  return identity
}

function responser(mime, render) {
  return (request, response, data) => {
    const body = render(data) // security header for content sniffing

    response.setHeader('X-Content-Type-Options', 'nosniff') // standard headers

    response.setHeader('Content-Type', `${mime}; charset=${CHARSET}`)
    response.setHeader('Content-Length', Buffer.byteLength(body, CHARSET)) // body

    response.end(body, CHARSET)
  }
}

class ServeDirectory {
  constructor(root, options) {
    // root required
    // resolve root to absolute and normalize
    root = path__default['default'].resolve(root)
    this.root = root
    this.responser = {}
    this.options = {
      hidden: defaults.hidden,
      relative: defaults.relative,
    }
    this.imports = _objectSpread2({}, imports)
    this.config(defaults)

    if (options) {
      this.config(options)
    }
  }

  config({hidden, relative, imports = {}, process = []} = {}) {
    if (isBoolean(hidden)) {
      this.options.hidden = hidden
    }

    if (isBoolean(relative)) {
      this.options.relative = relative
    }

    Object.assign(this.imports, imports)

    var _iterator = _createForOfIteratorHelper(process.filter(Boolean)),
      _step

    try {
      for (_iterator.s(); !(_step = _iterator.n()).done; ) {
        let _step$value = _step.value,
          accept = _step$value.accept,
          render = _step$value.render
        accept = (Array.isArray(accept) ? accept : accept.split(','))
          .map((x) => x.trim())
          .filter(Boolean)

        var _iterator2 = _createForOfIteratorHelper(accept),
          _step2

        try {
          for (_iterator2.s(); !(_step2 = _iterator2.n()).done; ) {
            const type = _step2.value

            if (render) {
              this.responser[type] = responser(
                type,
                template(render, {
                  imports: this.imports,
                })
              )
            } else {
              delete this.responser[type]
            }
          }
        } catch (err) {
          _iterator2.e(err)
        } finally {
          _iterator2.f()
        }
      }
    } catch (err) {
      _iterator.e(err)
    } finally {
      _iterator.f()
    }
  }

  middleware(request, response, next) {
    try {
      return new Connection$1(this, request, response, next).process()
    } catch (error) {
      return next(error)
    }
  }
}

var Directory = ServeDirectory

function middleware(root, options) {
  const sd = new Directory(root, options)
  return sd.middleware.bind(sd)
}

module.exports = middleware
